package com.xpn.xwiki.xmlrpc3;

import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xmlrpc.XmlRpcException;
import org.xwiki.xmlrpc.Page;
import org.xwiki.xmlrpc.PageSummary;
import org.xwiki.xmlrpc.Space;
import org.xwiki.xmlrpc.SpaceSummary;
import org.xwiki.xmlrpc.XWikiObject;
import org.xwiki.xmlrpc.XmlRpcUtils;

import com.xpn.xwiki.XWikiContext;
import com.xpn.xwiki.XWikiException;
import com.xpn.xwiki.api.Document;
import com.xpn.xwiki.api.Element;
import com.xpn.xwiki.api.Property;
import com.xpn.xwiki.api.XWiki;
import com.xpn.xwiki.objects.classes.BaseClass;
import com.xpn.xwiki.web.XWikiRequest;
import com.xpn.xwiki.web.XWikiResponse;
import com.xpn.xwiki.web.XWikiServletContext;
import com.xpn.xwiki.web.XWikiServletRequest;

public class XWikiRpcHandler
{
    private XWikiRequest xwikiRequest;

    private XWikiResponse xwikiResponse;

    private XWikiServletContext xwikiServletContext;

    private static final Log log = LogFactory.getLog(XWikiRpcHandler.class);

    public void init(HttpServlet servlet, HttpServletRequest httpRequest)
    {
        xwikiRequest = new XWikiServletRequest(httpRequest);
        xwikiResponse = (XWikiResponse) XWikiUtils.mock(XWikiResponse.class);
        ServletContext servletContext = servlet.getServletContext();
        xwikiServletContext = new XWikiServletContext(servletContext);
    }

    /***********************************************************************************************
     * Authentication methods
     **********************************************************************************************/
    
    public String login(String userName, String password) throws XWikiException, XmlRpcException
    {
        XWikiContext context =
            XWikiUtils.getXWikiContext(xwikiRequest, xwikiResponse, xwikiServletContext);
        com.xpn.xwiki.XWiki xwiki = context.getWiki();
        String token;

        if (xwiki.getAuthService().authenticate(userName, password, context) != null) {
            // Generate "unique" token using a random number
            token = xwiki.generateValidationKey(128);
            String ip = context.getRequest().getRemoteAddr();

            XWikiUtils.getTokens(context).put(token,
                new XmlRpcUser(String.format("XWiki.%s", userName), ip));

            return token;
        } else {
            throw new XmlRpcException(String.format("Authentication failed for user %s.",
                userName));
        }
    }

    public Boolean logout(String token) throws XWikiException, XmlRpcException
    {
        XWikiContext context =
            XWikiUtils.getXWikiContext(xwikiRequest, xwikiResponse, xwikiServletContext);
        XWikiUtils.checkToken(token, context);

        return XWikiUtils.getTokens(context).remove(token) != null;
    }

    /***********************************************************************************************
     * Spaces
     **********************************************************************************************/
    
    public List getSpaces(String token) throws XWikiException, XmlRpcException
    {
        XWiki xwiki =
            XWikiUtils.checkTokenAndGetXWikiApiWrapper(token, xwikiRequest, xwikiResponse,
                xwikiServletContext);

        List result = new ArrayList();
        List<String> spaceNames = xwiki.getSpaces();
        for (String spaceName : spaceNames) {
            String spaceWebHome = String.format("%s.WebHome", spaceName);

            SpaceSummary spaceSummary = null;

            if (!xwiki.exists(spaceWebHome)) {
                spaceSummary = new SpaceSummary(spaceName, "", "global", "");
            } else {
                Document doc = xwiki.getDocument(spaceWebHome);

                /* If doc is null, then we don't have the rights to access the document */
                if (doc != null) {
                    spaceSummary =
                        new SpaceSummary(spaceName, doc.getTitle(), "global", doc.getURL("view"));
                }
            }

            /* Might be null if we don't have the rights to access the document */
            if (spaceSummary != null) {
                result.add(spaceSummary.toMap());
            }
        }

        return result;
    }

    public Object getSpace(String token, String spaceName) throws XWikiException, XmlRpcException
    {
        XWiki xwiki =
            XWikiUtils.checkTokenAndGetXWikiApiWrapper(token, xwikiRequest, xwikiResponse,
                xwikiServletContext);

        if (!xwiki.getSpaces().contains(spaceName)) {
            throw new XmlRpcException(String.format("Space '%s' does not exist", spaceName));
        }

        Space space = null;
        String spaceWebHome = String.format("%s.WebHome", spaceName);
        if (!xwiki.exists(spaceWebHome)) {
            space = new Space(spaceName, "", "", "global", "", "No description available");
        } else {
            Document doc = xwiki.getDocument(spaceWebHome);

            /* If doc is null, then we don't have the rights to access the document */
            if (doc != null) {
                space =
                    new Space(spaceName,
                        doc.getTitle(),
                        "global",
                        doc.getURL("view"),
                        spaceWebHome,
                        "No description available");
            } else {
                throw new XmlRpcException(String.format("Space '%s' cannot be accessed",
                    spaceName));
            }

        }

        return space.toMap();
    }

    /***********************************************************************************************
     * Pages
     **********************************************************************************************/
    
    public List getPages(String token, String spaceName) throws XWikiException, XmlRpcException
    {
        XWiki xwiki =
            XWikiUtils.checkTokenAndGetXWikiApiWrapper(token, xwikiRequest, xwikiResponse,
                xwikiServletContext);

        List result = new ArrayList();
        List<String> pageNames = xwiki.getSpaceDocsName(spaceName);
        for (String pageName : pageNames) {
            String pageFullName = String.format("%s.%s", spaceName, pageName);
            if (!xwiki.exists(pageFullName)) {
                log.warn(String.format(
                    "Page '%s' appears to be in space '%s' but no information is available.",
                    pageName));
            } else {
                Document doc = xwiki.getDocument(pageFullName);
                String pageTitle = doc.getTitle();
                if (pageTitle.equals("")) {
                    pageTitle = pageName;
                }
                PageSummary pageSummary =
                    new PageSummary(pageFullName, spaceName, doc.getParent(), pageTitle, doc
                        .getURL("view"), 0);

                result.add(pageSummary.toMap());
            }
        }

        return result;
    }

    public Object getPage(String token, String pageId) throws XWikiException, XmlRpcException
    {
        XWiki xwiki =
            XWikiUtils.checkTokenAndGetXWikiApiWrapper(token, xwikiRequest, xwikiResponse,
                xwikiServletContext);

        if (!xwiki.exists(pageId)) {
            throw new XmlRpcException(String.format("Unable to get page %s", pageId));
        }

        Document doc = xwiki.getDocument(pageId);
        String pageTitle = doc.getTitle();
        if (pageTitle.equals("")) {
            pageTitle = doc.getName();
        }

        Page page =
            new Page(pageId,
                doc.getSpace(),
                doc.getParent(),
                pageTitle,
                doc.getURL(),
                0,
                doc.getRCSVersion().at(0),
                doc.getContent(),
                doc.getCreationDate(),
                doc.getCreator(),
                doc.getContentUpdateDate(),
                doc.getContentAuthor(),
                doc.getName().equals("WebHome"),
                "current",
                true);

        return page.toMap();
    }

    public Object storePage(String token, Map pageMap) throws XWikiException, XmlRpcException
    {
        XWiki xwiki =
            XWikiUtils.checkTokenAndGetXWikiApiWrapper(token, xwikiRequest, xwikiResponse,
                xwikiServletContext);

        Document doc;
        Page page = new Page((Map<String, Object>) pageMap);
        doc = xwiki.getDocument(page.getId());

        doc.setContent(page.getContent());
        doc.setTitle(page.getTitle());
        doc.save();

        String pageTitle = doc.getTitle();
        if (pageTitle.equals("")) {
            pageTitle = doc.getName();
        }

        Page newPage =
            new Page(doc.getFullName(),
                doc.getSpace(),
                doc.getParent(),
                pageTitle,
                doc.getURL(),
                0,
                doc.getRCSVersion().at(0),
                doc.getContent(),
                doc.getCreationDate(),
                doc.getCreator(),
                doc.getContentUpdateDate(),
                doc.getContentAuthor(),
                doc.getName().equals("WebHome"),
                "current",
                true);

        return newPage.toMap();
    }

    /***********************************************************************************************
     * Objects
     **********************************************************************************************/
    
    public List getObjects(String token, String pageId) throws XWikiException, XmlRpcException
    {
        XWiki xwiki =
            XWikiUtils.checkTokenAndGetXWikiApiWrapper(token, xwikiRequest, xwikiResponse,
                xwikiServletContext);

        if (!xwiki.exists(pageId)) {
            throw new XmlRpcException(String.format("Unable to get page %s", pageId));
        }

        Document doc = xwiki.getDocument(pageId);

        List result = new ArrayList();

        Map<String, Vector<com.xpn.xwiki.api.Object>> classToObjectsMap = doc.getxWikiObjects();
        for (String className : classToObjectsMap.keySet()) {
            Vector<com.xpn.xwiki.api.Object> objects = classToObjectsMap.get(className);
            for (com.xpn.xwiki.api.Object object : objects) {
                Map<String, Object> fieldToValueMap = new HashMap<String, Object>();
                Map<String, Object> fieldToTypeMap = new HashMap<String, Object>();

                for (Object o : object.getProperties()) {
                    Property property = (Property) o;
                    String name = property.getName();
                    Object convertedValue = XmlRpcUtils.convertToXmlRpc(property.getValue());

                    fieldToValueMap.put(name, convertedValue);
                    fieldToTypeMap.put(name, convertedValue.getClass());
                }

                String prettyName = object.getPrettyName();
                if (prettyName == null || prettyName.equals("")) {
                    prettyName = String.format("%s[%d]", className, object.getNumber());
                }

                XWikiObject xwikiObject =
                    new XWikiObject(pageId,
                        className,
                        object.getNumber(),
                        prettyName,
                        fieldToTypeMap,
                        fieldToValueMap);

                result.add(xwikiObject.toMap());
            }
        }

        return result;
    }

    public Object storeObject(String token, Map objectMap) throws XWikiException, XmlRpcException
    {
        XWiki xwiki =
            XWikiUtils.checkTokenAndGetXWikiApiWrapper(token, xwikiRequest, xwikiResponse,
                xwikiServletContext);

        XWikiObject object = new XWikiObject((Map<String, Object>) objectMap);
        String pageId = object.getPageId();

        if (!xwiki.exists(pageId)) {
            throw new XmlRpcException(String.format("Unable to get page %s", pageId));
        }

        Document doc = xwiki.getDocument(pageId);

        com.xpn.xwiki.api.Object currentObject =
            doc.getObject(object.getClassName(), object.getId());

        for (Object o : currentObject.getProperties()) {
            Property property = (Property) o;
            String fieldName = property.getName();
            Object currentValue = currentObject.getProperty(fieldName).getValue();
            Object newValue = object.get(fieldName);
            Class currentValueType = currentValue.getClass();
            Class newValueType = newValue.getClass();

            /*
             * These are needed because Longs and TimeStamps are transfered as Integers and Dates.
             * So we have to convert them back to their actual types
             */
            if (currentValueType.equals(Long.class) && newValueType.equals(Integer.class)) {
                newValue = new Long(((Integer) newValue).intValue());
            }

            if (currentValueType.equals(Timestamp.class) && newValueType.equals(Date.class)) {
                newValue = new Timestamp(((Date) newValue).getTime());
            }

            currentObject.set(fieldName, newValue);

        }

        doc.save();

        return object.toMap();
    }
    
    public List getClasses(String token) throws XWikiException, XmlRpcException {
        //TODO
        return null;
    }
}
